#include <string.h>
#include <stdlib.h>

#include "line-wobbler-advance.h"
#include "graphics.h"

const u8* advance_level_names[3] = {"@ghaoyuxin", "@snakesandrews", "@switchstyle"};

int main()
{
    REG_DISPCNT = DCNT_MODE3 | DCNT_BG2;
    REG_SNDSTAT = SSTAT_ENABLE;
    REG_SNDDSCNT = SDS_DMG100;

    int y, x, ii, jj, kk, top, player_ii_shift, frame_ii = 0, player_cell_ii = PLAYER_START, accumulated_shift,
        attacking, enemy_count, player_view_offset = PLAYER_VIEW_OFFSET_MIN, player_cell_store,
        enemy_first_index, elapsed;
    u8 player_map_indicator_ii, player_map_indicator_offset = 4, spawn_count = 0, frame_modifier,
        mode = MODE_ORIGINAL, at_title = 0, title_cursor_ii = 0;
    u16 enemy_draw_length, spawn_timer, right, left;
    COLOR rgb;
    u8 component, paused = 0, start_released = 1, level_index = 0, arrow_released = 1, pause_index = 0;
    u16 wobble = 0x99CF;
    struct Enemy enemies[ENEMY_MAX_COUNT];
    u8 stream_count;
    struct Stream streams[64];
    u8 lava_count;
    struct Lava lava[16];
    COLOR palette[PALETTE_LENGTH] =
        {
            COLOR_RED,
            COLOR_ORANGE,
            COLOR_YELLOW,
            RGB15(15, 31, 0),
            COLOR_GREEN,
            RGB15(0, 31, 15),
            RGB15(0, 31, 31),
            RGB15(0, 15, 31),
            RGB15(0, 0, 31),
            RGB15(15, 0, 31),
            RGB15(31, 0, 31),
            RGB15(31, 0, 15)
        };
    u32 two_tone = (0xFFFF << 16) | RGB15(25, 25, 25);
    u8 goal_palette_index = 0;
    u8 suppressing_move;
    u8 player_intro_length;
    struct Boss boss;
    const u16 boss_spawn_intervals[] = {143, 125, 94};
    const u16 boss_positions[] = {432, 130, 502};
    u16 celebration = 0;
    u32 explosion = 0x5D36274D;
    u32 explosion_palette[32] = {
        0, 3, 11, 5, 9, 7, 2, 1, 5, 8, 1, 10, 4, 6, 1, 10,
        0, 9, 10, 5, 4, 8, 2, 7, 3, 6, 1, 9, 1, 11, 5, 8
    };
    u8 exploding = 0;
    int explosion_start = 0;
    u8 boop = 0;
    u8 booped = 0;
    int boop_start = 0;
    COLOR line[CELL_COUNT];
    u32 blank_line_32[CELL_COUNT / 2];
    u32* line_32 = (u32*) line;
    COLOR map[MAP_LENGTH];
    u32 blank_map_32[MAP_LENGTH / 2];
    u32* map_32 = (u32*) map;
    struct Enemy *e;
    struct Save save;

    // read save data
    bytecpy(&save, sram_mem + SAVE_OFFSET, sizeof(save));
    save.advance_unlocked = 1;
    if (save.advance_unlocked)
    {
        at_title = 1;
    }
    
    // hud
    blit(Best, TIME_DISPLAY_INDENT, RECORDS_TOP, BEST_WIDTH, BEST_HEIGHT, 0);
    right = TIME_DISPLAY_INDENT + BEST_WIDTH + HUD_SPACE + (GLYPH0_WIDTH * 5) + 6
        + GLYPHCOLON_WIDTH + GLYPHPERIOD_WIDTH + HUD_PADDING;
    top = RECORDS_TOP - HUD_PADDING;
    horizontal_line(0, top, right - 1, COLOR_WHITE);
    vertical_line(right, top + 1, SCREEN_HEIGHT - (top + 2), COLOR_WHITE);

    // blank line
    for (ii = 0; ii < CELL_COUNT / 2; ii++)
    {
        blank_line_32[ii] = two_tone;
    }

    // blank map
    for (ii = 0; ii < MAP_LENGTH / 2; ii++)
    {
        blank_map_32[ii] = 0xFFFFFFFF;
    }

    // draw title
    x = 6;
    top = 7;
    for (y = top; y < top + TITLE_HEIGHT; y++)
    {
        memcpy(&vid_mem[SCREEN_WIDTH * y + x], &Title[TITLE_WIDTH * (y - 7)], TITLE_WIDTH * 2);
    }

    // lava color stamp
    COLOR lava_colors[3][CELL_COUNT];
    for (ii = 0; ii < 3; ii++)
    {
        for (jj = 0; jj < CELL_COUNT; jj++)
        {
            if (ii == 0)
            {
                lava_colors[ii][jj] = LAVA_COLOR_SUNK;
            }
            else if (ii == 1)
            {
                lava_colors[ii][jj] = LAVA_COLOR_RISING;
            }
            else if (ii == 2)
            {
                lava_colors[ii][jj] = LAVA_COLOR_RISEN;
            }
        }
    }

    // stream color stamp
    COLOR stream_colors[CELL_COUNT + 256];
    for (ii = 0; ii < CELL_COUNT + 256; ii++)
    {
        component = 4 + (ii & 15);
        stream_colors[ii] = RGB15(component, component, component);
    }

    // enemy color stamp
    COLOR enemy_colors[CELL_COUNT];
    for (ii = 0; ii < CELL_COUNT; ii++)
    {
        enemy_colors[ii] = COLOR_RED;
    }

    // celebration color stamp
    COLOR celebration_colors[PALETTE_LENGTH][CELL_COUNT];
    for (ii = 0; ii < PALETTE_LENGTH; ii++)
    {
        for (jj = 0; jj < CELL_COUNT; jj++)
        {
            celebration_colors[ii][jj] = palette[ii];
        }
    }

    // introduction color stamp
    COLOR player_intro_colors[CELL_COUNT];
    for (ii = 0; ii < CELL_COUNT; ii++)
    {
        player_intro_colors[ii] = COLOR_GREEN;
    }

    // boss color stamp
    COLOR boss_colors[BOSS_LENGTH];
    for (ii = 0; ii < BOSS_LENGTH; ii++)
    {
        boss_colors[ii] = BOSS_COLOR;
    }

    // first screen
    if (!at_title)
    {
        reset_level(&player_cell_ii, &enemies, &enemy_count, &streams, &stream_count, &lava,
                    &lava_count, &suppressing_move, &player_intro_length, &level_index, 0, &spawn_timer,
                    &enemy_first_index, &player_view_offset, &boss, &save, elapsed, mode);
        draw_best_time(save.best_time_original);
        draw_current_time_border();
        draw_current_time(elapsed);
        draw_level_border();
    }
    else
    {
        draw_title_menu_options();
        draw_title_menu_cursor(title_cursor_ii);
        draw_best_time(save.best_time_original);
    }

    while (1)
    {
        // wait for vblank
        vid_vsync();

        if (at_title)
        {
            if (start_released && (KEY_DOWN_NOW(KEY_START) || KEY_DOWN_NOW(KEY_A)))
            {
                start_released = 0;
                if (title_cursor_ii == 0)
                {
                    mode = MODE_ORIGINAL;
                }
                else
                {
                    mode = MODE_ADVANCE;
                }
                erase(0, TITLE_MENU_TOP - 4, SCREEN_WIDTH, TITLE_MENU_CURSOR_TOP + CURSOR_HEIGHT - (TITLE_MENU_TOP - 4));
                reset_level(&player_cell_ii, &enemies, &enemy_count, &streams, &stream_count, &lava,
                            &lava_count, &suppressing_move, &player_intro_length, &level_index, 0, &spawn_timer,
                            &enemy_first_index, &player_view_offset, &boss, &save, elapsed, mode);
                draw_current_time_border();
                draw_level_border();
                at_title = 0;
                REG_SND1CNT = SSQR_ENV_BUILD(14, 0, 4) | SSQR_DUTY3_4;
                REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                REG_SND1SWEEP = SSW_DEC | SSW_SHIFT(0) | SSW_TIME(0);
                REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, 1);
            }
            else
            {
                if (arrow_released && (KEY_DOWN_NOW(KEY_LEFT) || KEY_DOWN_NOW(KEY_RIGHT)))
                {
                    title_cursor_ii = !title_cursor_ii;
                    draw_title_menu_cursor(title_cursor_ii);
                    if (title_cursor_ii == 0)
                    {
                        draw_best_time(save.best_time_original);
                        erase(TITLE_MENU_CURSOR_ADVANCE_LEFT, TITLE_MENU_CURSOR_TOP,
                              CURSOR_WIDTH, CURSOR_HEIGHT);
                    }
                    else
                    {
                        draw_best_time(save.best_time_advance);
                        erase(TITLE_MENU_CURSOR_ORIGINAL_LEFT, TITLE_MENU_CURSOR_TOP,
                              CURSOR_WIDTH, CURSOR_HEIGHT);
                    }
                    arrow_released = 0;
                    REG_SND1CNT = SSQR_ENV_BUILD(14, 0, 1) | SSQR_DUTY3_4;
                    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                    REG_SND1SWEEP = SSW_DEC | SSW_SHIFT(2) | SSW_TIME(0);
                    REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, 1);
                }
                if (!KEY_DOWN_NOW(KEY_LEFT) && !KEY_DOWN_NOW(KEY_RIGHT))
                {
                    arrow_released = 1;
                }
                if (!KEY_DOWN_NOW(KEY_START))
                {
                    start_released = 1;
                }
            }
        }
        else if (!paused)
        {
            // set attack state
            if (!suppressing_move && boss.phase <= 2)
            {
                attacking = KEY_DOWN_NOW(KEY_A);
            }
        
            // erase line
            memcpy(line_32, blank_line_32, CELL_COUNT * 2);

            // erase map
            memcpy(map_32, blank_map_32, MAP_LENGTH * 2);

            // update stream
            player_ii_shift = 0;
            for (ii = 0; ii < stream_count; ii++)
            {
                memcpy(&line[streams[ii].start],
                       &stream_colors[streams[ii].offset],
                       (streams[ii].end - streams[ii].start + 1) * 2);
                memcpy(&map[streams[ii].start / 3], &stream_colors[streams[ii].offset],
                       (streams[ii].end - streams[ii].start + 1) / 3 * 2);
                if (streams[ii].update_mod == 3 && !(frame_ii & 1))
                {
                    streams[ii].offset -= streams[ii].step;
                }
                else if (!(frame_ii & streams[ii].update_mod))
                {
                    streams[ii].offset -= streams[ii].step * 2;
                }
                if (!exploding && !player_intro_length && streams[ii].start <= player_cell_ii &&
                    streams[ii].end >= player_cell_ii && !(frame_ii & streams[ii].update_mod))
                {
                    player_ii_shift = streams[ii].step;
                }
            }

            // update lava
            if (!exploding)
            {
                for (ii = 0; ii < lava_count; ii++)
                {
                    if (lava[ii].time < lava[ii].split)
                    {
                        lava[ii].color_ii = 0;
                    }
                    else if (lava[ii].time < (lava[ii].split << 1))
                    {
                        lava[ii].color_ii = 1;
                    }
                    else
                    {
                        lava[ii].color_ii = 2;
                        if (lava[ii].start <= player_cell_ii && lava[ii].end >= player_cell_ii && !celebration)
                        {
                            explode(&exploding, &explosion_start, frame_ii);
                            break;
                        }
                    }
                    if (lava[ii].time++ > lava[ii].cycle)
                    {
                        lava[ii].time = 0;
                    }
                    if (lava[ii].color_ii && !exploding)
                    {
                        memcpy(&line[lava[ii].start], lava_colors[lava[ii].color_ii],
                               (lava[ii].end - lava[ii].start + 1) * 2);
                        memcpy(&map[lava[ii].start / 3], lava_colors[lava[ii].color_ii],
                               (lava[ii].end - lava[ii].start + 1) / 3 * 2);
                    }
                }
            }

            // wobble
            if (attacking)
            {
                jj = player_cell_ii + ATTACK_REACH;
                for (ii = player_cell_ii - ATTACK_REACH; ii < jj; ii++)
                {
                    if (ii >= 0 && ii < CELL_COUNT)
                    {
                        if (!(ii == player_cell_ii) && ((1 << (abs(ii - player_cell_ii) - 1)) & wobble))
                        {
                            line[ii] = COLOR_YELLOW;
                        }
                    }
                }
                jj = player_cell_ii / 3 + MAP_ATTACK_REACH - (frame_ii & 1) + 1;
                for (ii = player_cell_ii / 3 - MAP_ATTACK_REACH + (frame_ii & 1); ii < jj; ii++)
                {
                    if (ii >= 0 && ii < MAP_LENGTH)
                    {
                        if (ii != player_cell_ii / 3)
                        {
                            map[ii] = COLOR_YELLOW;
                        }
                    }
                }
            }

            // yay
            if (celebration > 0)
            {
                memcpy(&line[CELL_COUNT - celebration], celebration_colors[goal_palette_index],
                       celebration * 2);
                memcpy(&map[(CELL_COUNT - celebration) / 3], celebration_colors[goal_palette_index],
                       ((celebration - 1) / 3 + 1) * 2);
                celebration += GOAL_CELEBRATION_STEP;
                if (player_cell_ii > CELL_COUNT - 80 && !(frame_ii & 1))
                {
                    player_cell_ii -= GOAL_CELEBRATION_STEP;
                }
                if (celebration > GOAL_CELEBRATION_MAX)
                {
                    celebration = 0;
                    if (level_index == 8)
                    {
                        elapsed = 0;
                        at_title = 1;
                    }
                    else
                    {
                        reset_level(&player_cell_ii, &enemies, &enemy_count, &streams, &stream_count,
                                    &lava, &lava_count, &suppressing_move, &player_intro_length,
                                    &level_index, 1, &spawn_timer, &enemy_first_index,
                                    &player_view_offset, &boss, &save, elapsed, mode);
                    }
                }
            }
            else
            {
                memcpy(&line[GOAL_START], celebration_colors[goal_palette_index], GOAL_LENGTH * 2);
                memcpy(&map[GOAL_START / 3], celebration_colors[goal_palette_index], (GOAL_LENGTH / 3 + 1) * 2);
                if (player_cell_ii >= GOAL_START && (level_index != 8 || boss.phase > 2))
                {
                    celebration = GOAL_CELEBRATION_MIN;
                    attacking = 0;
                    suppressing_move = 1;
                    REG_SND1CNT = SSQR_ENV_BUILD(12, 1, 12) | SSQR_DUTY1_2;
                    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                    REG_SND1SWEEP = SSW_INC | SSW_SHIFT(7) | SSW_TIME(2);
                    REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, -2);
                }
            }

            // move player
            player_cell_store = player_cell_ii;
            player_cell_ii += player_ii_shift;
            if (!attacking && !exploding)
            {
                int moving = 0;
                if (KEY_DOWN_NOW(KEY_RIGHT))
                {
                    if (!suppressing_move)
                    {
                        player_cell_ii++;
                        if (player_cell_ii >= CELL_COUNT)
                        {
                            player_cell_ii = CELL_COUNT - 1;
                        }
                        moving = 1;
                    }
                }
                else if (KEY_DOWN_NOW(KEY_LEFT))
                {
                    if (!suppressing_move)
                    {
                        player_cell_ii--;
                        if (player_cell_ii < 0)
                        {
                            player_cell_ii = 0;
                        }
                        moving = 1;
                    }
                }
                else
                {
                    if (player_intro_length <= 0 && celebration < GOAL_CELEBRATION_MIN)
                    {
                        suppressing_move = 0;
                    }
                }
                if (moving)
                {
                    if (!booped && !(frame_ii % 5))
                    {
                        REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 1) | SSQR_DUTY1_8 | SSQR_LEN(3);
                        REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                        REG_SND1SWEEP = SSW_INC | SSW_SHIFT(7) | SSW_TIME(2);
                        REG_SND1FREQ = SFREQ_RESET | SFREQ_TIMED | SND_RATE(NOTE_E, 2);
                    }
                }
            }
            accumulated_shift += player_cell_store - player_cell_ii;
            if (frame_ii & 1)
            {
                player_view_offset += accumulated_shift >> 1;
                if (player_view_offset < PLAYER_VIEW_OFFSET_MIN)
                {
                    player_view_offset = PLAYER_VIEW_OFFSET_MIN;
                }
                else if (player_view_offset > PLAYER_VIEW_OFFSET_MAX)
                {
                    player_view_offset = PLAYER_VIEW_OFFSET_MAX;
                }
                accumulated_shift = 0;
            }

            // boss
            if (((mode == MODE_ORIGINAL && level_index == 8) ||
                 (mode == MODE_ADVANCE && level_index == 3)) && !celebration)
            {
                if (boss.phase > 2)
                {
                    frame_modifier = frame_ii % 96;
                    for (ii = 0; ii <= 52; ii++)
                    {
                        rgb = palette[((96 - frame_modifier + ii) % 96) / 8];
                        if (player_cell_ii + ii < CELL_COUNT)
                        {
                            line[player_cell_ii + ii] = rgb;
                            map[(player_cell_ii + ii) / 3] = rgb;
                        }
                        if (player_cell_ii - ii >= 0)
                        {
                            line[player_cell_ii - ii] = rgb;
                            map[(player_cell_ii - ii) / 3] = rgb;
                        }
                    }
                }
                else
                {
                    if (boss.moving)
                    {
                        boss.head += BOSS_STEP * boss.moving;
                        boss.tail += BOSS_STEP * boss.moving;
                        if (abs(boss.head - boss_positions[boss.phase]) <= BOSS_STEP)
                        {
                            boss.head = boss_positions[boss.phase];
                            boss.tail = boss.head + BOSS_LENGTH;
                            boss.moving = 0;
                        }
                    }
                    else if (attacking &&
                             (player_cell_ii > boss.head && player_cell_ii - ATTACK_REACH < boss.tail) ||
                             (player_cell_ii < boss.tail && player_cell_ii + ATTACK_REACH > boss.head))
                    {
                        if (boss.phase == 0)
                        {
                            boss.moving = -1;
                        }
                        else if (boss.phase == 1)
                        {
                            boss.moving = 1;
                        }
                        else
                        {
                            for (ii = 0; ii < enemy_count; ii++)
                            {
                                enemies[ii].alive = 0;
                            }
                        }
                        if (++boss.phase > 2)
                        {
                            save.advance_unlocked = 1;
                            if (mode == MODE_ORIGINAL)
                            {
                                if (elapsed < save.best_time_original || !save.best_time_original)
                                {
                                    save.best_time_original = elapsed;
                                }
                                draw_best_time(save.best_time_original);
                            }
                            else if (mode == MODE_ADVANCE)
                            {
                                if (elapsed < save.best_time_advance || !save.best_time_advance)
                                {
                                    save.best_time_advance = elapsed;
                                }
                                draw_best_time(save.best_time_advance);
                            }
                            bytecpy(sram_mem + SAVE_OFFSET, &save, sizeof(save));
                            draw_current_time(elapsed);
                        }
                        booped = 1;
                        boop_start = frame_ii;
                        REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 7) | SSQR_DUTY1_2;
                        REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                        REG_SND1SWEEP = SSW_DEC | SSW_SHIFT(7) | SSW_TIME(3);
                        REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, 1);
                    }
                    else if (spawn_timer++ > boss_spawn_intervals[boss.phase])
                    {
                        spawn_count = 0;
                        for (ii = 0; ii < enemy_count; ii++)
                        {
                            e = &enemies[ii];
                            if (!e->alive || e->head < 0 || e->tail >= CELL_COUNT)
                            {
                                if (spawn_count++ == 0)
                                {
                                    *e = (struct Enemy)
                                        {boss.head, 0, boss.head, boss.head, 3, 2, 0, 1, -1};
                                }
                                else
                                {
                                    *e = (struct Enemy)
                                        {boss.tail, CELL_COUNT, boss.tail, boss.tail, 3, 2, 0, 1, 1};
                                    break;
                                }
                            }
                        }
                        spawn_timer = 0;
                    }
                    memcpy(&line[boss.head], boss_colors, BOSS_LENGTH * 2);
                    memcpy(&map[boss.head / 3], boss_colors, (BOSS_LENGTH / 3 + 1) * 2);
                }
            }

            // update enemies
            if (player_intro_length == 0 && celebration == 0)
            {
                for (ii = 0; ii < enemy_count; ii++)
                {
                    e = &enemies[ii];
                    if (e->alive && !exploding)
                    {
                        // die
                        boop = 0;
                        if (attacking && abs(e->head - player_cell_ii) < ATTACK_REACH)
                        {
                            if (player_cell_ii < e->head && e->head <= player_cell_ii + ATTACK_REACH)
                            {
                                e->head = player_cell_ii + ATTACK_REACH + 1;
                                boop = 1;
                            }
                            else if (player_cell_ii > e->tail && e->tail >= player_cell_ii - ATTACK_REACH)
                            {
                                e->tail = player_cell_ii - ATTACK_REACH - 1;
                                boop = 1;
                            }
                            if (e->head > e->tail)
                            {
                                e->alive = 0;
                            }
                            if (boop && !booped)
                            {
                                booped = 1;
                                boop_start = frame_ii;
                                REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 7) | SSQR_DUTY3_4;
                                REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                                REG_SND1SWEEP = SSW_DEC | SSW_SHIFT(7) | SSW_TIME(3);
                                REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, 3);
                            }
                            if (!e->alive)
                            {
                                continue;
                            }
                        }
                        // collision
                        else if ((player_cell_ii >= e->head && player_cell_store <= e->head) ||
                                 (player_cell_ii <= e->tail && player_cell_store >= e->tail))
                        {
                            e->alive = 0;
                            explode(&exploding, &explosion_start, frame_ii);
                            break;
                        }
                        // draw
                        if (e->head < CELL_COUNT && e->tail >= 0)
                        {
                            if (e->head >= 0 && e->tail < CELL_COUNT)
                            {
                                enemy_draw_length = e->tail - e->head + 1;
                            }
                            else if (e->head < 0 && e->tail >= 0)
                            {
                                enemy_draw_length = e->tail + 1;
                            }
                            else if (e->tail >= CELL_COUNT)
                            {
                                enemy_draw_length = CELL_COUNT - e->head;
                            }
                            memcpy(&line[e->head], enemy_colors, enemy_draw_length * 2);
                            memcpy(&map[e->head / 3], enemy_colors, ((enemy_draw_length - 1) / 3 + 1) * 2);
                        }
                        // move
                        if (e->frequency && !(frame_ii % e->frequency))
                        {
                            if (e->head > player_cell_ii && e->head + e->step * e->direction <= player_cell_ii)
                            {
                                e->alive = 0;
                                explode(&exploding, &explosion_start, frame_ii);
                                break;
                            }
                            else if (e->tail < player_cell_ii && e->tail + e->step * e->direction >= player_cell_ii)
                            {
                                e->alive = 0;
                                explode(&exploding, &explosion_start, frame_ii);
                                break;
                            }
                            e->head += e->step * e->direction;
                            e->tail += e->step * e->direction;
                            if (e->loop && (e->head < e->start || e->tail > e->end))
                            {
                                e->direction *= -1;
                            }
                        }
                    }
                }
            }

            // introduction
            if (player_intro_length > 0)
            {
                memcpy(&line[player_cell_ii], player_intro_colors, player_intro_length * 2);
                memcpy(&map[player_map_indicator_offset], player_intro_colors, player_intro_length / 3 * 2);
                player_intro_length--;
                if (player_intro_length <= 0)
                {
                    suppressing_move = 0;
                    REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 7) | SSQR_DUTY1_2;
                    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                    REG_SND1SWEEP = SSW_INC | SSW_SHIFT(7) | SSW_TIME(4);
                    REG_SND1FREQ = SFREQ_RESET | SND_RATE(NOTE_C, 3);
                }
            }

            // player
            if (!attacking && !celebration && boss.phase <= 2)
            {
                line[player_cell_ii] = COLOR_GREEN;
                if (player_cell_ii / 3 < player_map_indicator_offset)
                {
                    player_map_indicator_ii = player_map_indicator_offset;
                }
                else if (player_cell_ii / 3 > MAP_LENGTH - player_map_indicator_offset)
                {
                    player_map_indicator_ii = MAP_LENGTH - player_map_indicator_offset;
                }
                else
                {
                    player_map_indicator_ii = player_cell_ii / 3;
                }
                map[player_map_indicator_ii] = COLOR_GREEN;
            }

            // update explosion
            if (exploding)
            {
                memset(map, explosion, MAP_LENGTH * 2);
                if (frame_ii - explosion_start >= EXPLOSION_LENGTH)
                {
                    exploding = 0;
                    reset_level(&player_cell_ii, &enemies, &enemy_count, &streams, &stream_count,
                                &lava, &lava_count, &suppressing_move, &player_intro_length,
                                &level_index, 0, &spawn_timer, &enemy_first_index, &player_view_offset,
                                &boss, &save, elapsed, mode);
                }
                else if (!(frame_ii & 3))
                {
                    explosion = rotl32(explosion, 1);
                }
            }

            // draw line
            jj = player_cell_ii - player_view_offset + CELLS_PER_SCREEN;
            for (ii = player_cell_ii - player_view_offset, x = 0; ii < jj;
                 ii++, x += CELL_STEP)
            {
                // outside line
                if (ii < 0 || ii >= CELL_COUNT)
                {
                    rgb = COLOR_BLACK;
                }
                // boom
                else if (exploding)
                {
                    u8 index = abs(ii - player_cell_ii) & 31;
                    if ((1 << (index - 1)) & explosion)
                    {
                        rgb = palette[explosion_palette[index]];
                    }
                    else
                    {
                        rgb = COLOR_WHITE;
                    }
                }
                else
                {
                    rgb = line[ii];
                }
                draw_cell(x, rgb);
            }

            // draw map
            memcpy(&vid_mem[SCREEN_WIDTH * MAP_TOP], map, MAP_LENGTH * 2);

            // clear & draw cursor
            x = player_cell_ii / 3 - (CURSOR_WIDTH >> 1) - 2;
            top = CURSOR_TOP;
            for (y = top; y < top + CURSOR_HEIGHT; y++)
            {
                if (x < 0)
                {
                    x = 0;
                }
                else if (x >= SCREEN_WIDTH - CURSOR_WIDTH - 2 || celebration > 0)
                {
                    x = SCREEN_WIDTH - CURSOR_WIDTH - 2;
                }
                memset(&vid_mem[SCREEN_WIDTH * y + x], 0, 4);
                memset(&vid_mem[SCREEN_WIDTH * y + x + CURSOR_WIDTH + 2], 0, 4);
                memcpy(&vid_mem[SCREEN_WIDTH * y + x + 2],
                       &Cursor[CURSOR_WIDTH * (y - top)], CURSOR_WIDTH * 2);
            }

            // update boop status
            if (booped && frame_ii - boop_start > FPS / 6)
            {
                booped = 0;
            }

            // animate attack
            if (!exploding && attacking && !(frame_ii & 1))
            {
                wobble = rotl16(wobble, 1);
                if (!booped && !(frame_ii & 3))
                {
                    REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 5) | SSQR_DUTY3_4 | SSQR_LEN(1);
                    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                    REG_SND1SWEEP = SSW_INC | SSW_SHIFT(3) | SSW_TIME(2);
                    REG_SND1FREQ = SFREQ_RESET | SFREQ_TIMED | SND_RATE(NOTE_A, -2);
                }
            }

            // animate goal
            goal_palette_index += GOAL_PALETTE_STEP;
            if (goal_palette_index > PALETTE_LENGTH)
            {
                goal_palette_index -= PALETTE_LENGTH;
            }

            // level 6 spawn
            if (mode == MODE_ORIGINAL && level_index == 4 && enemy_count < ENEMY_MAX_COUNT - 1 &&
                (spawn_timer > 0 || enemies[1].head < CELL_COUNT))
            {
                if (spawn_timer++ >= SPAWN_INTERVAL_LVL6)
                {
                    enemies[enemy_count++] = (struct Enemy)
                        {CELL_COUNT - 1, 0, CELL_COUNT - 1, CELL_COUNT - 1, 3, 3, 0, 1, -1};
                    enemies[enemy_count++] = (struct Enemy)
                        {0, CELL_COUNT, 0, 0, 3, 3, 0, 1, 1};
                    spawn_timer = 1;
                }
            }

            // advance mode rules
            if (mode == MODE_ADVANCE)
            {
                if (level_index == 2)
                {
                    if (spawn_timer++ >= FPS - 10 && enemy_count < ENEMY_MAX_COUNT)
                    {
                        enemies[enemy_count++] = (struct Enemy)
                            {CELL_COUNT - 1, 0, CELL_COUNT - 1, CELL_COUNT - 1, 1, 2, 0, 1, -1};
                        spawn_timer = 0;
                    }
                }
                else if (level_index == 1)
                {
                    for (ii = 0; ii < enemy_count; ii++)
                    {
                        if (!enemies[ii].alive && !(frame_ii % enemies[ii].frequency))
                        {
                            jj = enemies[ii].step * enemies[ii].direction;
                            enemies[ii].head = enemies[ii].tail;
                            enemies[ii].head += jj;
                            enemies[ii].tail += jj;
                        }
                        if (ii < 5 || (ii > 6 && ii < 12))
                        {
                            if (enemies[ii].head < enemies[ii].end)
                            {
                                enemies[ii].head += 170;
                                enemies[ii].tail += 170;
                                enemies[ii].alive = 1;
                            }
                        }
                        else
                        {
                            if (enemies[ii].head > enemies[ii].end)
                            {
                                enemies[ii].head -= 170;
                                enemies[ii].tail -= 170;
                                enemies[ii].alive = 1;
                            }
                        }
                    }
                }
            }

            // pause
            if (start_released && KEY_DOWN_NOW(KEY_START))
            {
                paused = 1;
                pause_index = 0;
                start_released = 0;
                REG_SNDDMGCNT = 0;
                for (y = CURSOR_TOP; y < CURSOR_TOP + CURSOR_HEIGHT; y++)
                {
                    memset(&vid_mem_32[SCREEN_WIDTH * y / 2], 0, SCREEN_WIDTH * 2);
                }
                for (y = PAUSE_MENU_TOP; y < PAUSE_MENU_TOP + RESUME_HEIGHT; y++)
                {
                    memcpy(&vid_mem[SCREEN_WIDTH * y + PAUSE_RESUME_LEFT],
                           &Resume[RESUME_WIDTH * (y - PAUSE_MENU_TOP)],
                           RESUME_WIDTH * 2);
                    memcpy(&vid_mem[SCREEN_WIDTH * y + PAUSE_RESET_LEFT],
                           &Reset[RESET_WIDTH * (y - PAUSE_MENU_TOP)],
                           RESET_WIDTH * 2);
                }
                REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 1) | SSQR_DUTY3_4;
                REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                REG_SND1SWEEP = SSW_DEC | SSW_SHIFT(6) | SSW_TIME(2);
                REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, 0);
            }
            else if (!KEY_DOWN_NOW(KEY_START))
            {
                start_released = 1;
            }

            // time elapsed
            elapsed += !celebration && !exploding && !player_intro_length;

            // draw title screen
            if (at_title)
            {
                erase_level();
                erase_current_time();
                erase_line();
                draw_title_menu_options();
                title_cursor_ii = 0;
                draw_title_menu_cursor(title_cursor_ii);
            }

            // increment frame index
            frame_ii++;
        }
        else
        {
            // unpause
            if (start_released && (KEY_DOWN_NOW(KEY_START) || KEY_DOWN_NOW(KEY_A)))
            {
                paused = 0;
                start_released = 0;
                erase(PAUSE_RESET_LEFT, PAUSE_MENU_TOP, RESET_WIDTH, RESET_HEIGHT);
                erase(PAUSE_RESUME_LEFT, PAUSE_MENU_TOP, RESUME_WIDTH, RESUME_HEIGHT);
                if (pause_index == 1)
                {
                    erase(PAUSE_CURSOR_RESET_LEFT, PAUSE_CURSOR_TOP, CURSOR_WIDTH, CURSOR_HEIGHT);
                    celebration = 0;
                    exploding = 0;
                    level_index = 0;
                    elapsed = 0;
                    if (save.advance_unlocked)
                    {
                        at_title = 1;
                        erase_line();
                        erase_level();
                        erase_current_time();
                        title_cursor_ii = 0;
                        draw_best_time(save.best_time_original);
                        draw_title_menu_options();
                        draw_title_menu_cursor(title_cursor_ii);
                    }
                    else
                    {
                        reset_level(&player_cell_ii, &enemies, &enemy_count, &streams, &stream_count, &lava,
                                    &lava_count, &suppressing_move, &player_intro_length, &level_index, 0, &spawn_timer,
                                    &enemy_first_index, &player_view_offset, &boss, &save, elapsed, mode);
                    }
                }
                else
                {
                    erase(PAUSE_CURSOR_RESUME_LEFT, PAUSE_CURSOR_TOP, CURSOR_WIDTH, CURSOR_HEIGHT);
                    REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 1) | SSQR_DUTY3_4;
                    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                    REG_SND1SWEEP = SSW_DEC | SSW_SHIFT(6) | SSW_TIME(2);
                    REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, 0);
                }
            }
            // move cursor
            else
            {
                x = (pause_index == 0) ? PAUSE_CURSOR_RESUME_LEFT : PAUSE_CURSOR_RESET_LEFT;
                for (y = PAUSE_CURSOR_TOP; y < PAUSE_CURSOR_TOP + CURSOR_HEIGHT; y++)
                {
                    memcpy(&vid_mem[SCREEN_WIDTH * y + x],
                           &Cursor[CURSOR_WIDTH * (y - PAUSE_CURSOR_TOP)],
                           CURSOR_WIDTH * 2);
                }
                if (arrow_released && (KEY_DOWN_NOW(KEY_RIGHT) || KEY_DOWN_NOW(KEY_LEFT)))
                {
                    memset(&vid_mem[SCREEN_WIDTH * PAUSE_CURSOR_TOP], 0,
                           SCREEN_WIDTH * 2 * CURSOR_HEIGHT);
                    pause_index = !pause_index;
                    arrow_released = 0;
                    REG_SND1CNT = SSQR_ENV_BUILD(14, 0, 1) | SSQR_DUTY3_4;
                    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
                    REG_SND1SWEEP = SSW_DEC | SSW_SHIFT(2) | SSW_TIME(0);
                    REG_SND1FREQ = SFREQ_RESET | SFREQ_HOLD | SND_RATE(NOTE_E, 1);
                }
                if (!KEY_DOWN_NOW(KEY_START))
                {
                    start_released = 1;
                }
                if (!KEY_DOWN_NOW(KEY_LEFT) && !KEY_DOWN_NOW(KEY_RIGHT))
                {
                    arrow_released = 1;
                }
            }
        }
    }

    return 0;
}

void blit(const u16* surface, u16 left, u16 top, u16 width, u16 height, u16 invert)
{
    int y, limit = top + height;
    if (!invert)
    {
        for (y = top; y < limit; y++)
        {
            memcpy(&vid_mem[SCREEN_WIDTH * y + left], &surface[width * (y - top)], width * 2);
        }
    }
    else
    {
        u16 *dst = &vid_mem[SCREEN_WIDTH * top + left];
        int size, x;
        for (size = width * height, x = 0; size; size--)
        {
            *dst++ = *surface++ ^ invert;
            if (x++ == width - 1)
            {
                x = 0;
                dst += SCREEN_WIDTH - width;
            }
        }
    }
}

void erase(u16 left, u16 top, u16 width, u16 height)
{
    int y, limit = top + height;
    for (y = top; y < limit; y++)
    {
        memset(&vid_mem[SCREEN_WIDTH * y + left], 0, width * 2);
    }
}

void explode(u8* exploding, int* start, int frame_ii)
{
    *exploding = 1;
    *start = frame_ii;
    REG_SND4CNT = SSQR_ENV_BUILD(12, 1, 7);
    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_NOISE, 7);
    REG_SND1SWEEP = SSW_OFF;
    REG_SND4FREQ = SFREQ_RESET | SND_RATE(NOTE_C, 2);
}

void reset_level(int* player_cell_ii, struct Enemy (*enemies)[256], int* enemy_count,
                 struct Stream (*streams)[64], u8* stream_count, struct Lava (*lava)[16],
                 u8* lava_count, u8* suppressing_move, u8* player_intro_length, u8* level_index,
                 u8 level_increment, u16* spawn_timer, int* enemy_first_index, int* player_view_offset,
                 struct Boss* boss, struct Save* save, int elapsed, u8 mode)
{
    memset(&vid_mem[SCREEN_WIDTH * CURSOR_TOP], 0, SCREEN_WIDTH * 4 * CURSOR_HEIGHT);
    int ii;
    *level_index += level_increment;
    if (mode == MODE_ORIGINAL)
    {
        if (*level_index == 0)
        {
            *enemy_count = 4;
            (*enemies)[0] = (struct Enemy) {117, 117, 117, 117, 0, 0, 1, 1, 1};
            (*enemies)[1] = (struct Enemy) {216, 236, 226, 226, 3, 1, 1, 1, -1};
            (*enemies)[2] = (struct Enemy) {360, 400, 380, 380, 3, 2, 1, 1, 1};
            (*enemies)[3] = (struct Enemy) {504, 524, 514, 514, 3, 4, 1, 1, -1};
            *stream_count = 0;
            *lava_count = 0;
        }
        else if (*level_index == 1)
        {
            *enemy_count = 1;
            (*enemies)[0] = (struct Enemy) {648, 649, 648, 649, 0, 0, 1, 1, 1};
            *stream_count = 2;
            (*streams)[0] = (struct Stream) {72, 431, 3, -1, 0};
            (*streams)[1] = (struct Stream) {504, 647, 1, 1, 0};
            *lava_count = 0;
        }
        else if (*level_index == 2)
        {
            *enemy_count = 3;
            (*enemies)[0] = (struct Enemy) {179, 0, 179, 195, 8, 1, 0, 1, -1};
            (*enemies)[1] = (struct Enemy) {395, 0, 395, 420, 5, 2, 0, 1, -1};
            (*enemies)[2] = (struct Enemy) {936, 0, 936, 1036, 5, 4, 0, 1, -1};
            *stream_count = 0;
            *lava_count = 0;
        }
        else if (*level_index == 3)
        {
            *enemy_count = 0;
            *stream_count = 2;
            (*streams)[0] = (struct Stream) {144, 215, 3, 1, 0};
            (*streams)[1] = (struct Stream) {491, 562, 3, -1, 0};
            *lava_count = 1;
            (*lava)[0] = (struct Lava) {216, 490, FPS * 8, 0, FPS * 8 / 3, 0};
        }
        else if (*level_index == 4)
        {
            *enemy_count = 3;
            (*enemies)[0] = (struct Enemy) {370, 370, 370, 370, 0, 0, 1, 1, 1};
            (*enemies)[1] = (struct Enemy) {1356, 0, 1356, 1356, 3, 3, 0, 1, -1};
            (*enemies)[2] = (struct Enemy) {-636, CELL_COUNT, -636, -636, 3, 3, 0, 1, 1};
            *stream_count = 0;
            *lava_count = 0;
        }
        else if (*level_index == 5)
        {
            *enemy_count = 1;
            (*enemies)[0] = (struct Enemy) {528, 547, 547, 547, 4, 1, 1, 1, -1};
            *stream_count = 0;
            *lava_count = 10;
            (*lava)[0] = (struct Lava) {100, 150, 456, 270, 456 / 3, 1};
            (*lava)[1] = (struct Lava) {150, 200, 456, 180, 456 / 3, 1};
            (*lava)[2] = (struct Lava) {200, 250, 456, 90, 456 / 3, 0};
            (*lava)[3] = (struct Lava) {250, 300, 456, 0, 456 / 3, 0};
            (*lava)[4] = (struct Lava) {300, 350, 456, 270, 456 / 3, 1};
            (*lava)[5] = (struct Lava) {350, 400, 456, 180, 456 / 3, 1};
            (*lava)[6] = (struct Lava) {400, 450, 456, 90, 456 / 3, 0};
            (*lava)[7] = (struct Lava) {450, 500, 456, 0, 456 / 3, 0};
            (*lava)[8] = (struct Lava) {500, 550, 456, 270, 456 / 3, 1};
            (*lava)[9] = (struct Lava) {550, 600, 456, 180, 456 / 3, 1};
        }
        else if (*level_index == 6)
        {
            *enemy_count = 30;
            u16 jj;
            for (ii = 0, jj = 251; ii < *enemy_count; jj += 30 - ii++)
            {
                (*enemies)[ii] = (struct Enemy) {jj, jj, jj, jj, 0, 0, 1, 1, 1};
            }
            *stream_count = 1;
            (*streams)[0] = (struct Stream) {179, 716, 3, 3, 0};
            *lava_count = 0;
        }
        else if (*level_index == 7)
        {
            *enemy_count = 6;
            (*enemies)[0] = (struct Enemy) {216, 260, 230, 230, 5, 3, 1, 1, 1};
            (*enemies)[1] = (struct Enemy) {288, 347, 300, 300, 4, 3, 1, 1, -1};
            (*enemies)[2] = (struct Enemy) {360, 406, 380, 380, 4, 3, 1, 1, 1};
            (*enemies)[3] = (struct Enemy) {432, 509, 480, 480, 5, 3, 1, 1, -1};
            (*enemies)[4] = (struct Enemy) {504, 570, 535, 535, 6, 5, 1, 1, 1};
            (*enemies)[5] = (struct Enemy) {576, 637, 600, 600, 1, 1, 1, 1, -1};
            *stream_count = 0;
            *lava_count = 4;
            (*lava)[0] = (struct Lava) {100, 169, 440, 147, 146, 1};
            (*lava)[1] = (struct Lava) {250, 334, 491, 327, 163, 2};
            (*lava)[2] = (struct Lava) {400, 494, 540, 0, 180, 0};
            (*lava)[3] = (struct Lava) {550, 619, 510, 341, 170, 2};
        }
        else if (*level_index == 8)
        {
            *enemy_count = 16;
            for (ii = 0; ii < *enemy_count; ii++)
            {
                (*enemies)[ii] = (struct Enemy) {0, 0, 0, 0, 0, 0, 0, 0, 0};
            }
            *stream_count = 0;
            *lava_count = 0;
        }
        boss->head = 432;
        boss->tail = boss->head + BOSS_LENGTH - 1;
        boss->phase = 0;
        boss->moving = 0;
    }
    else if (mode == MODE_ADVANCE)
    {
        // sandbox
        if (*level_index == 0)
        {
            /*
             - add objects (Enemy, Lava, Stream) using the syntax from other levels
             - object attributes are described in line-wobbler-advance.h
             - define the counts even if they are zero
             - if the game starts lagging, reduce the number of objects
             */
            *enemy_count = 0;
            *stream_count = 0;
            *lava_count = 0;
        }
        // yuxin
        else if (*level_index == 1)
        {
            *enemy_count = 14;
            (*enemies)[0] = (struct Enemy) {210, 40, 210, 210, 2, 1, 0, 1, -1};
            (*enemies)[1] = (struct Enemy) {210, 40, 176, 176, 2, 1, 0, 1, -1};
            (*enemies)[2] = (struct Enemy) {210, 40, 142, 142, 2, 1, 0, 1, -1};
            (*enemies)[3] = (struct Enemy) {210, 40, 108, 108, 2, 1, 0, 1, -1};
            (*enemies)[4] = (struct Enemy) {210, 40, 74, 74, 2, 1, 0, 1, -1};
            (*enemies)[5] = (struct Enemy) {210, 380, 210, 210, 1, 1, 0, 1, 1};
            (*enemies)[6] = (struct Enemy) {210, 380, 295, 295, 1, 1, 0, 1, 1};
            (*enemies)[7] = (struct Enemy) {550, 380, 550, 550, 3, 2, 0, 1, -1};
            (*enemies)[8] = (struct Enemy) {550, 380, 516, 516, 3, 2, 0, 1, -1};
            (*enemies)[9] = (struct Enemy) {550, 380, 482, 482, 3, 2, 0, 1, -1};
            (*enemies)[10] = (struct Enemy) {550, 380, 448, 448, 3, 2, 0, 1, -1};
            (*enemies)[11] = (struct Enemy) {550, 380, 414, 414, 3, 2, 0, 1, -1};
            (*enemies)[12] = (struct Enemy) {550, CELL_COUNT - 1, 550, 550, 1, 2, 0, 1, 1};
            (*enemies)[13] = (struct Enemy) {550, CELL_COUNT - 1, 635, 635, 1, 2, 0, 1, 1};
            *stream_count = 4;
            (*streams)[0] = (struct Stream) {40, 210, 3, -1, 0};
            (*streams)[1] = (struct Stream) {210, 380, 3, 1, 0};
            (*streams)[2] = (struct Stream) {380, 550, 1, -1, 0};
            (*streams)[3] = (struct Stream) {550, CELL_COUNT - 1, 3, 3, 0};
            *lava_count = 0;
        }
        // blake
        else if (*level_index == 2)
        {
            *enemy_count = 0;
            *stream_count = 0;
            *lava_count = 1;
            (*lava)[0] = (struct Lava) {0, CELL_COUNT - 4, FPS * 24, 0, FPS * 24 / 3, 0};
        }
        // shu
        else if (*level_index == 3)
        {
            *enemy_count = 0;
            *stream_count = 12;
            (*streams)[0]  = (struct Stream) {  0,  30, 1,  1, 0};
            (*streams)[1]  = (struct Stream) { 60,  90, 1,  1, 0};
            (*streams)[2]  = (struct Stream) {120, 150, 1,  1, 0};
            (*streams)[3]  = (struct Stream) {180, 210, 1,  1, 0};
            (*streams)[4]  = (struct Stream) {240, 270, 1,  1, 0};
            (*streams)[5]  = (struct Stream) {300, 330, 1,  1, 0};
            (*streams)[6]  = (struct Stream) {360, 390, 1,  1, 0};
            (*streams)[7]  = (struct Stream) {420, 450, 1,  1, 0};
            (*streams)[8]  = (struct Stream) {480, 510, 1,  1, 0};
            (*streams)[9]  = (struct Stream) {540, 570, 1,  1, 0};
            (*streams)[10] = (struct Stream) {600, 630, 1,  1, 0};
            (*streams)[11] = (struct Stream) {660, 690, 1, -1, 0};
            *lava_count = 12;
            (*lava)[0]  = (struct Lava) { 30,  60, 180,   0, 60, 0};
            (*lava)[1]  = (struct Lava) { 90, 120, 180,  60, 60, 0};
            (*lava)[2]  = (struct Lava) {150, 180, 180, 120, 60, 0};
            (*lava)[3]  = (struct Lava) {210, 240, 150,   0, 50, 0};
            (*lava)[4]  = (struct Lava) {270, 300, 150,  50, 50, 0};
            (*lava)[5]  = (struct Lava) {330, 360, 150, 100, 50, 0};
            (*lava)[6]  = (struct Lava) {390, 420, 120,   0, 40, 0};
            (*lava)[7]  = (struct Lava) {450, 480, 120,  40, 40, 0};
            (*lava)[8]  = (struct Lava) {510, 540, 120,  80, 40, 0};
            (*lava)[9]  = (struct Lava) {570, 600,  90,   0, 30, 0};
            (*lava)[10] = (struct Lava) {630, 660,  90,  30, 30, 0};
            (*lava)[11] = (struct Lava) {690, 716,  90,  60, 30, 0};
        }
        else if (*level_index == 4)
        {
            *enemy_count = 16;
            for (ii = 0; ii < *enemy_count; ii++)
            {
                (*enemies)[ii] = (struct Enemy) {0, 0, 0, 0, 0, 0, 0, 0, 0};
            }
            *stream_count = 0;
            *lava_count = 0;
        }
        boss->head = 432;
        boss->tail = boss->head + BOSS_LENGTH - 1;
        boss->phase = 0;
        boss->moving = 0;
    }
    blit(get_glyph(*level_index + 1), LEVEL_LEFT + LEVEL_WIDTH + HUD_SPACE, LEVEL_TOP,
         GLYPH0_WIDTH, GLYPH0_HEIGHT, 0);
    draw_current_time(elapsed);
    *player_cell_ii = PLAYER_START;
    *suppressing_move = 1;
    *player_intro_length = PLAYER_INTRO_LENGTH;
    *spawn_timer = 0;
    *enemy_first_index = 0;
    *player_view_offset = PLAYER_VIEW_OFFSET_MIN;
    REG_SND1CNT = SSQR_ENV_BUILD(12, 0, 7) | SSQR_DUTY1_2;
    REG_SNDDMGCNT = SDMG_BUILD_LR(SDMG_SQR1, 7);
    REG_SND1SWEEP = SSW_INC | SSW_SHIFT(7) | SSW_TIME(4);
    REG_SND1FREQ = SFREQ_RESET | 44400;
}

void draw_cell(int start, COLOR rgb)
{
    u32* dst = &vid_mem_32[(SCREEN_WIDTH * LINE_TOP + start) / 2];
    u32* end = &vid_mem_32[(SCREEN_WIDTH * LINE_BOTTOM + start) / 2];
    while (dst <= end)
    {
        *dst = rgb << 16 | rgb;
        dst += SCREEN_WIDTH / 2;
    }
}

u16 rotl16(u16 value, u8 count)
{
    return value << count | value >> (16 - count);
}

u32 rotl32(u32 value, u8 count)
{
    return value << count | value >> (32 - count);
}

/*
  by tepples & sgeos
  http://forum.gbadev.org/viewtopic.php?t=418
*/
void bytecpy(void* in_dst, const void* in_src, u32 length)
{
    u8* src = (u8*) in_src;
    u8* dst = (u8*) in_dst;
    for (; length > 0; length--)
    {
        *dst++ = *src++;
    }
}

void horizontal_line(u16 left, u16 top, u16 length, COLOR color)
{
    u16* dst = &vid_mem[SCREEN_WIDTH * top + left];
    u16* end = &vid_mem[SCREEN_WIDTH * top + left + length];
    while (dst <= end)
    {
        *dst++ = color;
    }
}

void vertical_line(u16 left, u16 top, u16 length, COLOR color)
{
    u16* dst = &vid_mem[SCREEN_WIDTH * top + left];
    u16* end = &vid_mem[SCREEN_WIDTH * (top + length) + left];
    while (dst <= end)
    {
        *dst = color;
        dst += SCREEN_WIDTH;
    }
}

void draw_best_time(int frame_count)
{
    struct Time time;
    parse_time(frame_count, &time);
    draw_time(&time, TIME_DISPLAY_INDENT + BEST_WIDTH + HUD_SPACE, RECORDS_TOP);
}

void draw_current_time_border()
{
    horizontal_line(CURRENT_TIME_LEFT + 1, RECORDS_BORDER_TOP,
                    SCREEN_WIDTH - CURRENT_TIME_LEFT - 2, COLOR_WHITE);
    vertical_line(CURRENT_TIME_LEFT, RECORDS_BORDER_TOP + 1,
                  SCREEN_HEIGHT - (RECORDS_BORDER_TOP + 2), COLOR_WHITE);
}

void draw_current_time(int elapsed)
{
    struct Time time;
    parse_time(elapsed, &time);
    draw_time(&time, SCREEN_WIDTH - TIME_DISPLAY_INDENT - (GLYPH0_WIDTH * 5 + 4) -
              (GLYPHCOLON_WIDTH + 1) - (GLYPHPERIOD_WIDTH + 1), RECORDS_TOP);
}

void erase_current_time()
{
    erase(CURRENT_TIME_LEFT, RECORDS_BORDER_TOP, SCREEN_WIDTH - CURRENT_TIME_LEFT,
          SCREEN_HEIGHT - RECORDS_BORDER_TOP);
}

void draw_level_border()
{
    blit(Level, LEVEL_LEFT, LEVEL_TOP, LEVEL_WIDTH, LEVEL_HEIGHT, 0);
    horizontal_line(LEVEL_BORDER_LEFT + 1, RECORDS_BORDER_TOP,
                    LEVEL_BORDER_RIGHT - LEVEL_BORDER_LEFT - 2, COLOR_WHITE);
    vertical_line(LEVEL_BORDER_LEFT, RECORDS_BORDER_TOP + 1,
                  SCREEN_HEIGHT - (RECORDS_BORDER_TOP + 2), COLOR_WHITE);
    vertical_line(LEVEL_BORDER_RIGHT, RECORDS_BORDER_TOP + 1,
                  SCREEN_HEIGHT - (RECORDS_BORDER_TOP + 2), COLOR_WHITE);
}

void erase_level()
{
    erase(LEVEL_BORDER_LEFT, RECORDS_BORDER_TOP, LEVEL_BORDER_RIGHT - LEVEL_BORDER_LEFT + 1,
          SCREEN_HEIGHT - RECORDS_BORDER_TOP);
}

void erase_line()
{
    erase(0, LINE_TOP, SCREEN_WIDTH, CURSOR_TOP + CURSOR_HEIGHT - LINE_TOP);
}

void draw_title_menu_options()
{
    horizontal_line(0, TITLE_MENU_TOP - 4, SCREEN_WIDTH - 1, COLOR_WHITE);
    horizontal_line(0, TITLE_MENU_TOP + GLYPH0_HEIGHT + 3, SCREEN_WIDTH - 1, COLOR_WHITE);
    blit(Original, TITLE_MENU_ORIGINAL_LEFT, TITLE_MENU_TOP, ORIGINAL_WIDTH, ORIGINAL_HEIGHT, 0);
    blit(Advance, TITLE_MENU_ADVANCE_LEFT, TITLE_MENU_TOP, ADVANCE_WIDTH, ADVANCE_HEIGHT, 0);
}

void draw_title_menu_cursor(u8 index)
{
    if (index == 0)
    {
        blit(Cursor, TITLE_MENU_CURSOR_ORIGINAL_LEFT, TITLE_MENU_CURSOR_TOP,
             CURSOR_WIDTH, CURSOR_HEIGHT, 0);
    }
    else
    {
        blit(Cursor, TITLE_MENU_CURSOR_ADVANCE_LEFT, TITLE_MENU_CURSOR_TOP,
             CURSOR_WIDTH, CURSOR_HEIGHT, 0);
    }
}

void parse_time(int count, struct Time* time)
{
    u8 remainder = count % FPS;
    time->ms = (remainder + remainder * 2 / 3) / 10;
    int seconds = count / 60;
    time->s = seconds % 60;
    time->min = seconds / 60;
    if (time->min > 99)
    {
        time->min = 99;
        time->s = 59;
        time->ms = 9;
    }
}

void draw_time(struct Time* time, int left, int top)
{
    draw_digits(time->min, left, top, time->min < 10);
    left += GLYPH0_WIDTH * 2 + 2;
    blit(GlyphCOLON, left, top, GLYPHCOLON_WIDTH, GLYPHCOLON_HEIGHT, 0);
    left += GLYPHCOLON_WIDTH + 1;
    draw_digits(time->s, left, top, time->s < 10);
    left += GLYPH0_WIDTH * 2 + 2;
    blit(GlyphPERIOD, left, top, GLYPHPERIOD_WIDTH, GLYPHPERIOD_HEIGHT, 0);
    left += GLYPHPERIOD_WIDTH + 1;
    draw_digits(time->ms, left, top, 0);
}

/*
  digit extraction by Britton Kerin
  http://stackoverflow.com/a/44537930/1256386
*/
void draw_digits(int number, int left, int top, int pad)
{
    int y, div = 1, digit_count = 1;
    const u16* glyph;
    while (div <= number / 10)
    {
        digit_count++;
        div *= 10;
    }
    while (pad + digit_count > 0)
    {
        if (pad > 0)
        {
            glyph = get_glyph(0);
            pad--;
        }
        else
        {
            glyph = get_glyph(number / div);
            number %= div;
            div /= 10;
            digit_count--;
        }
        blit(glyph, left, top, GLYPH0_WIDTH, GLYPH1_HEIGHT, 0);
        left += GLYPH0_WIDTH + 1;
    }
}

u8 count_digits(int number)
{
    number = abs(number);
    if (number == 0)
    {
        return 1;
    }
    u8 count = 0;
    while (number > 0)
    {
        number /= 10;
        count++;
    }
    return count;
}

const u16* get_glyph(u8 digit)
{
    switch (digit)
    {
    case 0:
        return Glyph0;
    case 1:
        return Glyph1;
    case 2:
        return Glyph2;
    case 3:
        return Glyph3;
    case 4:
        return Glyph4;
    case 5:
        return Glyph5;
    case 6:
        return Glyph6;
    case 7:
        return Glyph7;
    case 8:
        return Glyph8;
    case 9:
        return Glyph9;
    }
}
